(**********************************************************************)
(*                                                                    *)
(*                        CPA beats ooCFA                             *)
(*                                                                    *)
(*       Frédéric Besson,      Inria Rennes - Bretagne Atlantique     *)
(*                                                                    *)
(*  Copyright Institut National de Recherche en Informatique et en    *)
(*  Automatique. All rights reserved                                  *)
(*                                                                    *)
(**********************************************************************)
Require Import List.
Require Import Utf8.
Require Import StdNotations.
Require Import Fix.
Require Import Omega.

(** Properties about finite types *)

Section FType.
  
  (** Consider a type [T] inhabited by finitely many elements *)
  Hypothesis T : Type.
  Hypothesis t0 : T.
  Hypothesis AllT : list T.
  Hypothesis FinT : forall x : T, In x AllT.

  Lemma indT : forall P, (∀ x, In x AllT ->  P x) -> ∀ x, P x.
  (* begin hide *)
  Proof.
    intros.
    apply X. apply FinT.
  Qed.
  (* end hide *)

  Lemma invT : forall P, (∀ x, P x) -> (∀ x, In x AllT ->  P x).
  (* begin hide *)
  Proof.
    auto.
  Qed.
  (* end hide *)

  Section Max.

    Require Import Max.

    Hypothesis P : nat -> T -> Prop.
    Hypothesis P_all : forall r, exists n, P n r .
    Hypothesis P_mon : forall n n', n <= n' -> P n ⊆ P n'.

    Inductive forallP : list T -> nat -> Prop :=
    | forallNil : forall n r, P n r -> forallP  (r::nil) n
    | forallCons  : forall r n l n', P n r  -> forallP l n' -> forallP (r::l) (max n n').

    Lemma forallP_ex : forall l, l <> nil -> exists n, forallP l n.
    (* begin hide *)
    Proof.
      intros.
      induction l. congruence.
      destruct l.
      destruct (P_all a).
      exists x. constructor.
      auto.
      assert (t ::l <> nil) by congruence.
      destruct (IHl H0).
      destruct (P_all a).
      exists (max x0 x).
      constructor; auto.
    Qed.
    (* end hide *)

    Lemma forallP_mon : forall l n, forallP l n -> forall n', n <= n' -> forallP l n'.
    (* begin hide *)
    Proof.
      intros until n.
      intro.
      induction H.
      intros.
      constructor.  eapply P_mon ;eauto.
      intros.
      replace n'0  with (max n n'0).
      eapply forallCons ; eauto.
      apply IHforallP.
      generalize (le_max_r n n' ). omega.
      generalize (max_l n n').
      generalize (max_r n n').
      generalize (max_l n n'0).
      generalize (max_r n n'0).
      omega.
    Qed.
    (* end hide *)

    Lemma ex_swap : exists n, forall r, P n r.
    (* begin hide *)
    Proof.
    intros.
    assert (AllT <> nil).
     generalize (FinT t0).
     destruct AllT.
     simpl. tauto. congruence.
    destruct (forallP_ex _ H) ; auto.
    exists x.
    clear FinT.
    apply indT.
    induction AllT.
    simpl. tauto.
    intros.
    inv H0. simpl in H1.
    intuition congruence.
    intros.
    simpl in H1 ; destruct H1.
    subst.
    eapply P_mon with (2:= H4); eauto.
    apply le_max_l.
    destruct l; simpl in *.
    tauto.
    apply IHl ; auto. congruence.
    eapply forallP_mon ;eauto.
    apply le_max_r.
  Qed.
  (* end hide *)

End  Max.

End FType.

(* Local Variables: *)
(* coding: utf-8 *)
(* coq-prog-name: "/usr/local/bin/coqtop" *)
(* End: *)
